import pytest

from paten import Paten
from paten.error import ArgumentNameInvalidError


def test_http_trigger():
    app = Paten("pytest")

    @app.http_trigger("req", methods=["GET"], route="/pytest")
    @app.out_queue("arg_name", queue_name="example-queue")
    @app.out_http()
    def example(req, arg_name):
        pass


def test_http_trigger_name_error():
    app = Paten("pytest")

    with pytest.raises(ArgumentNameInvalidError):
        @app.http_trigger("req", methods=["GET"], route="/pytest")
        @app.out_http()
        def example(req_different_name):
            pass

    with pytest.raises(ArgumentNameInvalidError):
        @app.http_trigger("req", methods=["GET"], route="/pytest")
        @app.out_queue("arg_name", queue_name="example-queue")
        @app.out_http()
        def example(req, arg_name_different):
            pass



from inspect import signature
import json
import os
from typing import Union, Optional
from shutil import copyfile

from .error import ArgumentNameInvalidError


class Paten:

    def __init__(self, function_app_name: str, requirements: Optional[list] = None):
        self.function_app_name = function_app_name
        # manage function list including bindings
        self.function_info_list = []
        # manage bindings
        self.function_bind_list = []
        # requirements
        self.requirements = requirements

    def http_trigger(self, name, methods: Union[list, str], route: str, auth_level: str = "function"):
        def _wrapper(function):
            # check arguments
            sig = signature(function)
            if name not in sig.parameters:
                raise ArgumentNameInvalidError(f"{name} not in {function.__name__}")

            self.function_bind_list.append({
                "function_name": str(function.__name__),
                "values": {
                    "authLevel": auth_level,
                    "type": "httpTrigger",
                    "direction": "in",
                    "name": name,
                    "route": route,
                    "methods": methods
                }
            })

            self.function_info_list.append({
                "function_name": str(function.__name__),
                "function_json": {
                    "scriptFile": "__init__.py",
                    "entryPoint": str(function.__name__),
                    "bindings": [d['values'] for d in self.function_bind_list if
                                 d['function_name'] == str(function.__name__)]
                }
            })
            return function
        return _wrapper

    def out_http(self, name: Optional[str] = None):
        def _wrapper(function):
            _name = name if name is not None else "$return"

            self.function_bind_list.append({
                "function_name": str(function.__name__),
                "values": {
                    "type": "http",
                    "direction": "out",
                    "name": _name
                }
            })
            return function
        return _wrapper

    def out_queue(self, name: str, queue_name: str, connection: Optional[str] = None):
        def _wrapper(function):
            # check arguments
            sig = signature(function)
            if name not in sig.parameters:
                raise ArgumentNameInvalidError(f"{name} not in {function.__name__}")

            _connection = connection if connection is not None else "AzureWebJobsStorage"

            self.function_bind_list.append({
                "function_name": str(function.__name__),
                "values": {
                    "type": "queue",
                    "direction": "out",
                    "name": name,
                    "queueName": queue_name,
                    "connection": _connection
                }
            })
            return function
        return _wrapper

    @staticmethod
    def _generate_function_json():
        return {
            "scriptFile": "__init__.py"
        }

    def _generate_requirements_txt(self) -> list:
        if self.requirements is None:
            return []
        return self.requirements

    @staticmethod
    def _generate_host_json() -> dict:
        return {
            "version": "2.0",
            "extensionBundle": {
                "id": "Microsoft.Azure.Functions.ExtensionBundle",
                "version": "[1.*, 2.0.0)"
            }
        }

    @staticmethod
    def _generate_local_settings_json() -> dict:
        return {
            "IsEncrypted": False,
            "Values": {
                "AzureWebJobsStorage": "",
                "FUNCTIONS_WORKER_RUNTIME": "python"
            }
        }

    @staticmethod
    def _generate_proxies_json() -> dict:
        return {
            "$schema": "http://json.schemastore.org/proxies",
            "proxies": {}
        }

    def export(self):
        output_dir = f"./.paten"
        os.makedirs(output_dir, exist_ok=True)

        # requirementsファイルの複製
        copyfile("requirements.txt", f"./.paten/requirements.txt")

        with open("./.paten/proxies.json", "w") as f:
            json.dump(self._generate_proxies_json(), f)

        with open("./.paten/local.settings.json", "w") as f:
            json.dump(self._generate_local_settings_json(), f)

        with open("./.paten/host.json", "w") as f:
            json.dump(self._generate_host_json(), f)

        with open("./.paten/requirements.txt", mode='w') as f:
            f.writelines('\n'.join(self._generate_requirements_txt()))

        for func in self.function_info_list:
            output_dir = f"./.paten/{func['function_name']}"
            os.makedirs(output_dir, exist_ok=True)

            # function.jsonの取得と配置
            out: dict = func['function_json']
            with open(f'{output_dir}/function.json', 'w') as f:
                json.dump(out, f)

            # 関数ファイルの配置
            copyfile("app.py", f"{output_dir}/__init__.py")

    def plan(self) -> list:
        output_list = ["app.py", "|"]
        for func in self.function_info_list:
            output_list.append(f"|-{func['function_name']}")
            for bindings in func['function_json']['bindings']:
                output_list.append(f"|  |-[{bindings['type']}] {bindings['name']}")
            output_list.append("|")

        return output_list



import os

import click

from paten.constants import TEMPLATE_APP, GITIGNORE, WELCOME_PROMPT
from .factory import CliFactory


@click.group(invoke_without_command=True)
@click.option('--function-app-dir')
@click.pass_context
def cmd(ctx, function_app_dir: str):
    if function_app_dir is None:
        function_app_dir = os.getcwd()
    if ctx.invoked_subcommand is None:
        click.echo('This is parent!')
    ctx.obj['function_app_dir'] = function_app_dir
    ctx.obj['factory'] = CliFactory(function_app_dir=function_app_dir)


@cmd.command("deploy")
@click.pass_context
def deploy(ctx):
    click.echo('start to deploy')
    cli_factory: CliFactory = ctx.obj['factory']
    paten_app = cli_factory.load_paten_app()
    paten_app.export()


@cmd.command("build")
@click.pass_context
def build(ctx):
    cli_factory: CliFactory = ctx.obj['factory']
    paten_app = cli_factory.load_paten_app()
    paten_app.export()

    output_list = paten_app.plan()
    for s in output_list:
        click.echo(s)

    click.echo("\nyour function app is successfully built to `./.paten/*`")


@cmd.command("local")
@click.option('--host', default='127.0.0.1')
@click.option('--port', default=8000, type=click.INT)
@click.pass_context
def local(ctx, host='127.0.0.1', port=8000):
    click.echo(f'hosting at local at {host}:{port}')

    cli_factory: CliFactory = ctx.obj['factory']
    paten_app = cli_factory.load_paten_app()


@cmd.command("plan")
@click.pass_context
def plan(ctx):
    cli_factory: CliFactory = ctx.obj['factory']
    paten_app = cli_factory.load_paten_app()
    output_list = paten_app.plan()
    for s in output_list:
        click.echo(s)

    click.echo("\ntype `paten build` to generate files to `./.paten/*`")


def _create_new_project_skeleton(function_app_name: str):
    paten_dir = os.path.join(function_app_name, '.paten')
    os.makedirs(paten_dir)
    with open(os.path.join(function_app_name, 'requirements.txt'), 'w'):
        pass
    with open(os.path.join(function_app_name, 'app.py'), 'w') as f:
        f.write(TEMPLATE_APP % function_app_name)
    with open(os.path.join(function_app_name, '.gitignore'), 'w') as f:
        f.write(GITIGNORE)


@cmd.command("new-project")
@click.argument('function_app_name')
def new_project(function_app_name: str):
    click.echo(WELCOME_PROMPT % function_app_name)
    _create_new_project_skeleton(function_app_name=function_app_name)


def main():
    # コンテキストから参照するアトリビュートを渡す
    cmd(obj={})


if __name__ == '__main__':
    main()



import os
import sys

# テスト対象のファイルへのパスを通している
# pytestの設定
PARENT_PATH = os.path.abspath(os.path.dirname(os.path.abspath(__file__)))
SOURCE_PATH = PARENT_PATH.rsplit('/', 1)[0]

sys.path.append(f"{SOURCE_PATH}")



import importlib
import sys

from paten import Paten


class CliFactory:

    def __init__(self, function_app_dir: str):
        self.function_app_dir = function_app_dir

    def load_paten_app(self) -> Paten:
        if self.function_app_dir not in sys.path:
            sys.path.insert(0, self.function_app_dir)
        try:
            app = importlib.import_module('app')
            paten_app = getattr(app, 'app')
        except SyntaxError as e:
            message = (
                          'Unable to import your app.py file:\n\n'
                          'File "%s", line %s\n'
                          '  %s\n'
                          'SyntaxError: %s'
                      ) % (getattr(e, 'filename'), e.lineno, e.text, e.msg)
            raise RuntimeError(message)
        return paten_app



TEMPLATE_APP = """
import azure.functions as func
from paten import Paten

app = Paten('%s')


@app.http_trigger('req', methods=['GET'], route='/http')
@app.out_http()
def example_http_function(req: func.HttpRequest) -> func.HttpResponse:
    name = req.params.get('name')
    
    # response
    return func.HttpResponse(name)


@app.http_trigger('req', methods=['GET'], route='/queue')
@app.out_queue('your_queue', queue_name='example_queue')
@app.out_http()
def example_http_queue_function(req: func.HttpRequest, your_queue: func.Out[str]) -> func.HttpResponse:
    name = req.params.get('name')
    # set name in your queue
    your_queue.set(name)
    
    # response
    return func.HttpResponse(name)
    
"""


GITIGNORE = """
.paten/*
"""


WELCOME_PROMPT = r"""
 _____             _______  ______  _   _
|  __ \     /\    |__   __||  ____|| \ | |
| |__) |   /  \      | |   | |__   |  \| |
|  ___/   / /\ \     | |   |  __|  | . ` |
| |      / ____ \    | |   | |____ | |\  |
|_|     /_/    \_\   |_|   |______||_| \_|

Welcome PATEN.
Your function app `%s` is generated
"""